
from miniappi.core.stream.session import StreamSession
from .models import BaseMessage, BaseContent
from .message_types import InputMessage, PutRoot

from .context import user_context
from .utils.message import handle_message
import logging

logger = logging.getLogger(__name__)


class AppSession(StreamSession):

    def get_logger(self, name: str | None = None):
        return logger

    async def send(self, data):
        if isinstance(data, BaseContent):
            # Considering as put message
            data = PutRoot(
                data=data
            )
        if isinstance(data, dict):
            data = InputMessage(**data)
        if not isinstance(data, BaseMessage):
            raise TypeError(f"Expected: {BaseMessage!r}, given: {type(data)!r}")
        out = await super().send(data.model_dump(exclude_none=True))
        self._set_context(data)
        return out

    def _set_context(self, msg: InputMessage):
        ... # TODO
        #curr_content = channel_context.current_content
        # handle_message(curr_content, msg)



import logging

from miniappi.core.context import Context

from miniappi.core.stream.base import Streamer
from miniappi.core.stream.session import StreamSession
from .models import BaseMessage, BaseContent
from .message_types import InputMessage, PutRoot

from .context import user_context as default_channel_context, app_context as default_app_context
from .utils.message import handle_message
from .session import AppSession

logger = logging.getLogger(__name__)

class App(Streamer[AppSession]):
    """Miniappi application

    This class is the entrypoint
    to create apps and it handles
    the communication with Miniappi
    server.

    Args:
        user_context (Context, optional): 
            Optional extra user scoped context
            to keep track on user level
            data. This is automatically
            scoped and can be set globally.
        app_context (Context, optional):
            Optional app scoped context
            to keep track on app level data.
            This is automatically
            scoped and can be set globally.

    Examples:
        ```python
        from miniappi import App

        app = App()

        @app.on_open()
        async def new_user(session):
            print("New user joined")
            ...

        app.run()
        ```
    """
    cls_session = AppSession

    def __init__(self,
                 *,
                 user_context: Context | None = None,
                 app_context: Context | None = None):
        super().__init__(None)

        self.channel_context = user_context
        self.app_context = app_context

    def get_app_context_managers(self, args: dict):
        init_args = dict(
            app=self,
            sessions=self.sessions,
            **args
        )
        ctx = [
            *self.app_context_managers,
            default_app_context.enter(init_args),
        ]
        if self.app_context:
            ctx.append(self.app_context.enter())
        return ctx

    def get_channel_context_managers(self, session: AppSession):
        init_args = dict(
            session=session,
            request_id=session.start_args.request_id,
        )
        ctx = [
            *self.channel_context_managers,
            default_channel_context.enter(
                init_args
            ),
        ]
        if self.channel_context:
            ctx.append(self.channel_context.enter())
        return ctx

    def get_logger(self, name: str):
        return logger



from pathlib import Path
from pydantic_settings import BaseSettings, SettingsConfigDict

class Settings(BaseSettings):
    "API settings (by default dev)"
    model_config = SettingsConfigDict(
        env_prefix='python_api_'
    )

    log_dir: Path = Path().cwd() / "logs"

    url_start: str = "https://miniappi.com/api/v1/streams/apps/start"
    url_apps: str = "https://miniappi.com/apps"

    echo_url: bool | None = True

settings = Settings()



from typing import Generic, TypeVar, AsyncGenerator, Literal, Self, AsyncContextManager
import json
import logging
from dataclasses import dataclass
from contextlib import asynccontextmanager
from abc import ABC, abstractmethod
from httpx_ws import aconnect_ws, AsyncWebSocketSession
from httpx import AsyncClient
from pydantic import BaseModel

from miniappi.config import settings
from .exceptions import CloseSessionException

ConnectionT = TypeVar("ConnectionT")
SessionT = TypeVar("SessionT")

class BaseStartArgs(BaseModel):
    request_id: str

@dataclass
class Message:
    channel: str
    request_id: str | None
    data: dict

class AbstractConnection(ABC):

    @abstractmethod
    async def publish(self, data: dict):
        "Publish a message to the connection"
        ...

    @abstractmethod
    async def listen(self) -> AsyncGenerator[Message]:
        "Listen messages from the connection"
        ...

    @abstractmethod
    async def listen_start(self) -> AsyncGenerator[BaseStartArgs]:
        "Listen start message from the connection"
        ...

class AbstractChannel(ABC, Generic[ConnectionT]):

    def __init__(self):
        self.app_url: str | None = None
        self.app_name: str | None = None

    @abstractmethod
    def connect(self) -> AsyncContextManager[ConnectionT]:
        ...

class AbstractClient(ABC, Generic[SessionT]):

    @abstractmethod
    def from_start_args(self, args) -> SessionT:
        "Init client from start args (for session)"

    @abstractmethod
    def from_init_channel(self, app_name: str) -> SessionT:
        "Init client from channel name (for app init)"
        ...

# Websocket

class WebsocketStartArgs(BaseStartArgs):
    channel: str = None

class WebsocketConnection(AbstractConnection):

    def __init__(self, ws: AsyncWebSocketSession, client: "WebsocketClient"):
        self.ws = ws
        self.client = client

    async def publish(self, data: dict):
        data = json.dumps(data)
        await self.ws.send_text(data)

    async def listen(self):
        while True:
            data = await self.ws.receive_text()
            if data == "OFF":
                # Users disconnected
                raise CloseSessionException("User closed the session")
            message = json.loads(data)
            yield Message(
                channel=self.client.channel,
                request_id=self.client.request_id,
                data=message
            )

    async def listen_start(self):
        async for msg in self.listen():
            yield WebsocketStartArgs(**msg.data)

class WebsocketChannel(AbstractChannel[WebsocketConnection]):

    def __init__(self, client: AsyncClient, channel: str, request_id: str | None, is_anonymous: bool = True):
        self.client = client
        self.channel = channel
        self.request_id = request_id
        self.is_anonymous = is_anonymous

        self.app_url: str | None = None
        self.app_name: str | None = None
        super().__init__()

    @asynccontextmanager
    async def connect(self):
        logger = logging.getLogger(__name__)
        logger.info(f"Connecting: {self.channel}")
        async with aconnect_ws(self.channel, client=self.client, keepalive_ping_interval_seconds=200, keepalive_ping_timeout_seconds=200) as ws:
            is_start = self.request_id is None
            if is_start:
                is_anonymous = self.is_anonymous
                if is_anonymous:
                    # The start sends a payload for
                    # generated UUID
                    # We show this UUID to the user
                    # so they can go to the app page
                    uuid = await ws.receive_json()
                    self.app_name = uuid
                    self.app_url = f"{settings.url_apps}/{uuid}"
                else:
                    self.app_url = f"{settings.url_apps}/{self.app_name}"
            yield WebsocketConnection(ws, client=self)


class WebsocketClient(AbstractClient[WebsocketChannel]):

    def __init__(self, client: AsyncClient | None = None):
        self.client = client or AsyncClient()

    def from_start_args(self, args: WebsocketStartArgs) -> WebsocketChannel:
        return WebsocketChannel(
            client=self.client,
            channel=args.channel,
            request_id=args.request_id,
        )

    def from_init_channel(self, app_name: str | None) -> WebsocketChannel:
        is_anonymous = app_name is None
        url = (
            settings.url_start
            if is_anonymous
            else f"{settings.url_start}/{app_name}"
        )
        return WebsocketChannel(
            client=self.client,
            channel=url,
            request_id=None,
            is_anonymous=is_anonymous
        )
